/*
 * fifo.h
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 *****************************************************************************
 *
 * pozadavky:
 * 	- jeden odesilatel (funguje vice)
 * 	- vice prijemcu
 * 	- zarucit poradi
 *
 * 
 */
#ifndef JK_FIFO_H
#define JK_FIFO_H

#include <stdio.h>
#include <pthread.h>
#include <queue>

#define CLASS "Fifo"
#ifdef DEBUG
#	define FIFO_H__DEBUG DEBUG
#	undef DEBUG
#endif
#define DEBUG 0
#include "debug.h"

template <class T>
class Fifo {
	public:
		Fifo(void);
		~Fifo(void);

		// Add to end of fifo
		void push(T obj);

		// get from begin of fifo
		bool pop(T *obj);

		// return true if fifo is empty
		inline bool is_hungry();

		// (un)register source
		int source_ref();
		int source_unref();

		// (un)register target
		int target_ref();
		int target_unref();

	private:
		unsigned int targets_count, sources_count;
		pthread_mutex_t access_lock;
		pthread_cond_t wait_cond, empty_cond, full_cond;

		static const unsigned int fifo_max_size;
		std::queue<T> fifo;
		unsigned int pop_wait_counter;
};


/****************************************************************************/

/*
 * g++-3.3 -Wall -D_REENTRANT -DSOUND_OSS -DFORCE_DEBUG=-1  -c fifo.cc -o fifo.o
 * fifo.cc:26: warning: keyword `export' not implemented, and will be ignored
 */

/****************************************************************************/

template <class T>
const unsigned int Fifo<T>::fifo_max_size = 10;
		
template <class T>
Fifo<T>::Fifo(void) {
	DBG_MSG(8, "creating...");
	
	pthread_mutex_init(&access_lock, NULL);
	pthread_cond_init(&full_cond, NULL);
	pthread_cond_init(&empty_cond, NULL);
	pthread_cond_init(&wait_cond, NULL);
	
	targets_count = 0;
	sources_count = 1;
	pop_wait_counter = 0;

	DBG_MSG(8, "done");
}

/****************************************************************************/

template <class T>
Fifo<T>::~Fifo(void) {
	DBG_MSG(8, "removing...");
	pthread_mutex_destroy(&access_lock);
	pthread_cond_destroy(&full_cond);
	pthread_cond_destroy(&empty_cond);
	pthread_cond_destroy(&wait_cond);
	DBG_MSG(8, "done");
}

/****************************************************************************/

template <class T>
void Fifo<T>::push(T obj) {
	
	DBG_MSG(7, "access_lock...");
	pthread_mutex_lock(&access_lock);
	DBG_MSG(7, "access_lock = on");
	
	while (fifo.size() >= fifo_max_size) {
		DBG_MSG(6, "Fifo is full, waiting...");
		pthread_cond_wait(&full_cond, &access_lock);
	}
	
	fifo.push(obj);
	DBG_MSG(8, "objects left in fifo: %d", fifo.size());
	pthread_cond_broadcast(&empty_cond);
	
	DBG_MSG(7, "access_lock = off");
	pthread_mutex_unlock(&access_lock);
}

/****************************************************************************/

template <class T>
bool Fifo<T>::pop(T *obj) {
	DBG_MSG(7, "access_lock...");
	pthread_mutex_lock(&access_lock);
	DBG_MSG(7, "access_lock = on");
	
	while (fifo.empty()) {
		if (sources_count == 0 && fifo.empty()) {
			DBG_MSG(6, "Fifo is hungry.");
			DBG_MSG(7, "access_lock = off");
			pthread_mutex_unlock(&access_lock);
			return false;
		}
		DBG_MSG(6, "Fifo is empty, waiting...");
		pthread_cond_wait(&empty_cond, &access_lock);
	}
	
	// precteni dat
	DBG_MSG(6, "Reading data");
	*obj = fifo.front();
	DBG_MSG(8, "objects left in fifo: %d", fifo.size());

	if (++pop_wait_counter >= targets_count) {
		// jsme posledni -- likvidace objektu + uvolneni ostatnich
		DBG_MSG(6, "I'm last -- removing data");
		fifo.pop();
		pthread_cond_broadcast(&full_cond);

		pop_wait_counter = 0;
		pthread_cond_broadcast(&wait_cond);
	} else {
		// pockame na ostatni
		DBG_MSG(6, "Waiting for others");
		while (pthread_cond_wait(&wait_cond, &access_lock) != 0)
			;
		DBG_MSG(8, "Others are ready");
	}

	DBG_MSG(7, "access_lock = off");
	pthread_mutex_unlock(&access_lock);
	return true;
}

/****************************************************************************/

template <class T>
bool Fifo<T>::is_hungry() {
	return (sources_count == 0 && fifo.empty());
}

/****************************************************************************/

template <class T>
int Fifo<T>::source_ref() {
	DBG_MSG(8, "");
	return ++sources_count;
}

/****************************************************************************/

template <class T>
int Fifo<T>::source_unref() {
	DBG_MSG(8, "");
	if (sources_count) {
		sources_count--;
		pthread_cond_broadcast(&empty_cond);
		DBG_MSG(6, "Source unref");
		return sources_count;
	} else {
		DBG_MSG(1, "CRITICAL: sources_count == 0 !");
		return 0;
	}
}

/****************************************************************************/

template <class T>
int Fifo<T>::target_ref() {
	DBG_MSG(8, "");
	return ++targets_count;
}

/****************************************************************************/

template <class T>
int Fifo<T>::target_unref() {
	DBG_MSG(8, "");
	if (targets_count)
		return --targets_count;
	else {
		DBG_MSG(1, "CRITICAL: targets_count == 0 !");
		return 0;
	}
}

/****************************************************************************/


#undef DEBUG
#ifdef FIFO_H__DEBUG
#	define DEBUG FIFO_H__DEBUG
#	undef FIFO_H__DEBUG
#endif

#undef CLASS
#endif

