/*
 * game.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "game.h"
#include "game_object.h"
#include "game_world.h"
#include "sample_msg.h"

#define CLASS "Game"
#define DEBUG 4
#include "debug.h"

#define PRAH 		0.70	// cast z nejsilnejsi amplitudy
#define AVG_COUNT	40	// nejsilnejsich frekvenci v prumeru
#define AVG_MUL		4	// nasobek prumeru
#define MIN_INTENSITY	1800000	// minimalni intenzita zvuku

const Game::KeyBound Game::key_bound[KEY_NUM] = {
	{    3,    6, 'k' },	// krtek
	{    6,   10, 'K' },
	{   24,   34, 'z' },	// zahradnik
	{   36,   47, 'Z' }
};

Game::Game(Screen *screen, SampleFifo *fifo, int height, int width) {
	DBG_MSG(1, "Initializing...");
	
	this->screen = screen;
	this->main_done = 0;
	this->world_height = height;
	this->world_width = width;
	this->fifo = fifo;
	// this->objects;
	
	fifo->target_ref();
	
	DBG_MSG(2, "Done.");
}

Game::~Game(void) {
	DBG_MSG(1, "Cleaning up...");
	
	main_done = 1;
	
	DBG_MSG(2, "Done.");
}

void Game::main(void) {
	SampleMsg *msg;
	GameWorld *world;
	int time = 0;
	int key[KEY_NUM];
	int old_key[KEY_NUM];		// stare klavesy
#if DEBUG_LEVEL >= 6
	int int_min = -1, int_max = -1;
#endif

	const float freq_zoom = 4.0;

	DBG_MSG(1, "Started.");

	world = new GameWorld(world_width, world_height);
	screen->render_world(world);
	screen->set_status(3, 0, 3, 0);
	screen->draw_all();

	/* define freq rule */
	for (int b = 0; b < KEY_NUM; b++) {
		screen->add_freq(key_bound[b].min, key_bound[b].max, key_bound[b].symbol);
	}
	
	while (!main_done) {
		if (!fifo->pop(&msg)) {
			break;
		}
		if (msg == NULL) {
			/* timeout */
			time++;
			world->step();
			screen->render_world(world);
			screen->draw_all();
		} else {
			/* sample message */
#if DEBUG_LEVEL >= 6
			if (int_min != -1) {
				int_min = msg->intensity < int_min ? msg->intensity : int_min;
				int_max = msg->intensity > int_max ? msg->intensity : int_max;
			} else {
				int_min = msg->intensity;
				int_max = msg->intensity;
			}
#endif
			DBG_MSG(9, "Received top frequency %d nj", msg->data[0].freq);
			//DBG_MSG(6, "Total intensity = %8ld (min = %d, max = %d).",
			//		msg->intensity, int_min, int_max);
			sample2keys(key, msg, 1);
			screen->render_freq(msg->freq, (int)(msg->size / freq_zoom),
					freq_minval, msg->intensity, MIN_INTENSITY);
			screen->set_status(key[0] | (key[1] << 1), world->get_score_krtek(),
					key[2] | (key[3] << 1), world->get_score_zahradnik());
			screen->draw_freq();

			/* odstraneni dlouhych stisku */
			for (int k = 0; k < KEY_NUM; k++) {
				if (key[k]) {
					if (old_key[k])
						key[k] = 0;
					old_key[k] = 1;
				} else {
					old_key[k] = 0;
				}
			}
			world->interpret_keys(key);
			msg->unref();
			screen->flush();
		}
	}

	fifo->target_unref();

	DBG_MSG(1, "Exit.");
	return;
}

void Game::kill(void) {
	DBG_MSG(1, "Hey Game::main, you are dead!");
	main_done = 1;
}

/*****************************************************************************
 * Mstn metody
 ****************************************************************************/

/*****************************************************************************
 * Prevod frekvencniho spektra na klavesy (akce)
 *
 * Detekuje frekvence s velkou amplitudou a podle toho nastavuje flagy
 * jednotlivych pasem. Na zacatku je male pasmo, ktere se vynechava (nejnizsi
 * frekvence = zadny zvuk).
 *
 * TODO
 * 	v msg prijde i celkova intenzita sejmuteho kousku. Takze by to chtelo
 * 	ji pouzit.
 */
void Game::sample2keys(int key[KEY_NUM], SampleMsg *msg, int clear) {
	
	if (clear)
		bzero(key, sizeof(int) * KEY_NUM);

	/* pocitani prune hodnoty */
	int sum = 0, i;
	for (i = 0; i < msg->size && i < AVG_COUNT; i++) {
		sum += msg->data[i].val;
	}
	int avg = sum / i;

	/* urceni minimalni hodnoty */
	int minval_prah = (int)(msg->data[0].val * PRAH);
	int minval_avg = (int)(avg * AVG_MUL);
	freq_minval = minval_avg > minval_prah ? minval_avg : minval_prah;
	
	DBG_MSG(5, "minval_avg = %d; minval_prah = %d", minval_avg, minval_prah);
	
	/* pokud neni signal dostatesne silny, je na nic */
	if (msg->intensity < MIN_INTENSITY)
		return;
	
	/* scan spektra */
	for (i = 0; i < msg->size && msg->data[i].val > freq_minval; i++) {
		for (int b = 0; b < KEY_NUM; b++) {
			if (msg->data[i].freq > key_bound[b].min && msg->data[i].freq < key_bound[b].max) {
				key[b] = 1;
				break;
			}
		}
	}
}

#undef CLASS

